class Observer(object):
    def observe(self, name, *args, **kwargs):
        raise NotImplementedError()


class DispatchObserver(Observer):
    def __init__(self):
        self.observers = []

    def add_observer(self, observer):
        self.observers.append(observer)

    def observe(self, name, *args, **kwargs):
        for observer in self.observers:
            observer.observe(name, *args, **kwargs)


class SpecializedObserver(Observer):
    def observe(self, name, *args, **kwargs):
        instance_method_ref = getattr(self, "observe_{}".format(name))
        instance_method_ref(*args, **kwargs)



import random

import numpy as np

from pywmi.smt_check import evaluate
from pywmi.smt_print import pretty_print, pretty_print_instance
from typing import Tuple, List


class SelectionStrategy(object):
    def select_active(
        self, domain, data, labels, formula, active_indices
    ) -> Tuple[np.ndarray, np.ndarray, List[int]]:
        raise NotImplementedError()


class AllViolationsStrategy(SelectionStrategy):
    def select_active(
        self, domain, data, labels, formula, active_indices
    ) -> Tuple[np.ndarray, np.ndarray, List[int]]:
        active_set = set(active_indices)
        learned_labels = evaluate(domain, formula, data)
        differences = np.logical_xor(labels, learned_labels)
        difference_set = set(np.where(differences)[0])
        # print(active_set)
        # print(difference_set)
        # print(pretty_print(formula))
        # for i in active_set & difference_set:
        #     print(i)
        #     print(pretty_print_instance(domain, data[i]))
        #     print(labels[i], learned_labels[i])
        # print()
        # assert len(active_set & difference_set) == 0
        return data, labels, sorted(difference_set - active_set)


class RandomViolationsStrategy(AllViolationsStrategy):
    def __init__(self, sample_size):
        self.sample_size = sample_size
        self.last_violations = None

    def select_active(
        self, domain, data, labels, formula, active_indices
    ) -> Tuple[np.ndarray, np.ndarray, List[int]]:
        data, labels, all_violations = AllViolationsStrategy.select_active(
            self, domain, data, labels, formula, active_indices
        )
        self.last_violations = list(all_violations)
        sample_size = min(self.sample_size, len(self.last_violations))
        return data, labels, random.sample(self.last_violations, sample_size)


class WeightedRandomViolationsStrategy(AllViolationsStrategy):
    def __init__(self, sample_size, weights):
        self.sample_size = sample_size
        self.last_violations = None
        self.weights = weights

    def select_active(
        self, domain, data, labels, formula, active_indices
    ) -> Tuple[np.ndarray, np.ndarray, List[int]]:
        data, labels, all_violations = AllViolationsStrategy.select_active(
            self, domain, data, labels, formula, active_indices
        )
        self.last_violations = list(all_violations)
        sample_size = min(self.sample_size, len(self.last_violations))
        import sampling

        return (
            data,
            labels,
            sampling.sample_weighted(
                zip(
                    self.last_violations,
                    [self.weights[i] for i in self.last_violations],
                ),
                sample_size,
            ),
        )


class MaxViolationsStrategy(AllViolationsStrategy):
    def __init__(self, sample_size, weights):
        self.sample_size = sample_size
        self.last_violations = None
        self.weights = weights

    def select_active(
        self, domain, data, labels, formula, active_indices
    ) -> Tuple[np.ndarray, np.ndarray, List[int]]:
        all_violations = list(
            AllViolationsStrategy.select_active(
                self, domain, data, labels, formula, active_indices
            )
        )
        self.last_violations = all_violations
        sample_size = min(self.sample_size, len(all_violations))
        weighted_violations = zip(
            all_violations, [self.weights[i] for i in all_violations]
        )
        weighted_violations = sorted(weighted_violations, key=lambda t: t[1])
        # noinspection PyTypeChecker
        return data, labels, [t[0] for t in weighted_violations[0:sample_size]]



# Given a number of points:
# - Train a DT (scale points?)
# - For every point compute distance to the decision boundary
import sklearn.tree as tree

import pysmt.shortcuts as smt

from .core import MaxViolationsStrategy


class DecisionTreeSelection(MaxViolationsStrategy):
    def __init__(self):
        super().__init__(1, None)

    def select_active(self, domain, data, labels, formula, active_indices):
        if self.weights is None:
            self.weights = [
                min(d.values()) for d in get_distances(domain, data, labels)
            ]
        return super().select_active(domain, data, labels, formula, active_indices)


def convert(domain, data, labels):
    # def _convert(var, val):
    #     if domain.var_types[var] == smt.BOOL:
    #         return 1 if val else 0
    #     elif domain.var_types[var] == smt.REAL:
    #         return float(val)

    # feature_matrix = []
    # labels = []
    # for instance, label in data:
    #     feature_matrix.append([_convert(v, instance[v]) for v in domain.variables])
    #     labels.append(1 if label else 0)
    return data, labels


def learn_dt(feature_matrix, labels, **kwargs):
    # noinspection PyArgumentList
    estimator = tree.DecisionTreeClassifier(**kwargs)
    estimator.fit(feature_matrix, labels)
    return estimator


def export_dt(dt):
    import graphviz

    dot_data = tree.export_graphviz(dt, out_file=None)
    graph = graphviz.Source(dot_data)
    graph.render("DT")


def get_distances_dt(dt, domain, feature_matrix):
    # Include more features than trained with?

    leave_id = dt.apply(feature_matrix)
    feature = dt.tree_.feature
    threshold = dt.tree_.threshold
    node_indicator = dt.decision_path(feature_matrix)

    distances = []

    for sample_id in range(len(feature_matrix)):
        distance = dict()
        node_index = node_indicator.indices[
            node_indicator.indptr[sample_id] : node_indicator.indptr[sample_id + 1]
        ]
        for node_id in node_index:
            variable = domain.variables[feature[node_id]]
            if (
                leave_id[sample_id] != node_id
                and domain.var_types[variable] == smt.REAL
            ):
                new_distance = abs(
                    feature_matrix[sample_id][feature[node_id]] - threshold[node_id]
                )
                if variable not in distance or new_distance < distance[variable]:
                    distance[variable] = new_distance
        distances.append(distance)

    return distances


def get_distances(domain, data, labels):
    # feature_matrix, labels = convert(domain, data, labels)
    dt = learn_dt(data, labels)
    return get_distances_dt(dt, domain, data)


if __name__ == "__main__":
    pass



from __future__ import print_function, division

import json

from pywmi import smt_to_nested
from pywmi.smt_print import pretty_print

from incal.incremental_learner import IncrementalObserver


class LoggingObserver(IncrementalObserver):
    def __init__(self, filename, k, h, seed=None, verbose=True, violation_counter=None):
        self.filename = filename
        self.verbose = verbose
        self.violation_counter = violation_counter
        self.k = k
        self.h = h

        if filename is not None:
            with open(self.filename, "w") as f:
                print("", file=f, end="")

        if seed is not None:
            self.log({"type": "seed", "seed": seed, "k": self.k, "h": self.h})

    def log(self, flat):
        if self.filename is not None:
            with open(self.filename, "a") as f:
                print(json.dumps(flat), file=f)

    def observe_initial(self, data, labels, initial_indices):
        flat = {"type": "initial", "indices": initial_indices, "k": self.k, "h": self.h}
        if self.verbose:
            print("Starting with {} examples".format(len(initial_indices)))
        self.log(flat)

    def observe_iteration(
        self, data, labels, formula, new_active_indices, solving_time, selection_time
    ):
        flat = {
            "type": "update",
            "theory": smt_to_nested(formula),
            "indices": [int(v) for v in new_active_indices],
            "solving_time": solving_time,
            "selection_time": selection_time,
            "k": self.k,
            "h": self.h,
        }
        if self.violation_counter is not None:
            flat["violations"] = [
                int(v) for v in self.violation_counter.last_violations
            ]

        if self.verbose:
            print("Found model after {:.2f}s".format(solving_time))
            print(pretty_print(formula))
            if self.violation_counter is not None:
                violation_count = len(self.violation_counter.last_violations)
                selected_count = len(new_active_indices)
                print(
                    "Selected {} of {} violations in {:.2f}s".format(
                        selected_count, violation_count, selection_time
                    )
                )
        self.log(flat)



import random


# Implementation modified from https://stackoverflow.com/a/2149533/253387


class Node(object):
    # Each node in the heap has a weight, original weight, value, and total weight.
    # The total weight, self.tw, is self.w plus the weight of any children.
    __slots__ = ["w", "ow", "v", "tw"]

    def __init__(self, w, ow, v, tw):
        self.w, self.ow, self.v, self.tw = w, ow, v, tw


def rws_heap(items):
    # h is the heap. It's like a binary tree that lives in an array.
    # It has a Node for each pair in `items`. h[1] is the root. Each
    # other Node h[i] has a parent at h[i>>1]. Each node has up to 2
    # children, h[i<<1] and h[(i<<1)+1].  To get this nice simple
    # arithmetic, we have to leave h[0] vacant.
    h = [None]  # leave h[0] vacant
    for v, w in items:
        h.append(Node(w, w, v, w))
    for i in range(len(h) - 1, 1, -1):  # total up the tws
        h[i >> 1].tw += h[i].tw  # add h[i]'s total to its parent
    return h


def rws_heap_pop(h):
    gas = h[1].tw * random.random()  # start with a random amount of gas

    i = 1  # start driving at the root
    while gas >= h[i].w:  # while we have enough gas to get past node i:
        gas -= h[i].w  #   drive past node i
        i <<= 1  #   move to first child
        if gas >= h[i].tw:  #   if we have enough gas:
            gas -= h[i].tw  #     drive past first child and descendants
            i += 1  #     move to second child
    w = h[i].w  # out of gas! h[i] is the selected node.
    v = h[i].v

    h[i].w = 0  # make sure this node isn't chosen again
    while i:  # fix up total weights
        h[i].tw -= w
        i >>= 1
    return v


def sample_weighted(items, n):  # items are pairs (value, weight)
    heap = rws_heap(items)  # just make a heap...
    for i in range(n):
        yield rws_heap_pop(heap)  # and pop n items off it.



from pysmt.shortcuts import Real, Plus, Symbol, And, GE
from pysmt.typing import REAL

from .incremental_learner import IncrementalLearner
from .learner import NoFormulaFound

try:
    import gurobipy as gurobi
except ImportError:
    gurobi = None


class LpLearnerMilp(IncrementalLearner):
    def __init__(
        self, half_space_count, selection_strategy, allow_negations=True, symmetries=""
    ):
        IncrementalLearner.__init__(self, "cnf_smt", selection_strategy, False)
        self.half_space_count = half_space_count
        self.allow_negations = allow_negations
        self.symmetries = symmetries
        self.active_indices = set()
        # self.smt_solver = smt_solver

    def learn_partial(self, solver, domain, data, labels, new_active_indices):
        if gurobi is None:
            raise RuntimeError("Gurobi is not installed.")

        # noinspection PyPep8Naming
        GRB = gurobi.GRB
        quicksum = gurobi.quicksum

        m = gurobi.Model("milp")
        # m.setParam('TimeLimit', 60*90)
        m.setParam("OutputFlag", False)

        # constants
        n_c = self.half_space_count
        n_v = len(domain.real_vars)  # number of variables #j
        bigm = 10000000
        ep = 1
        wmax = 1000
        cmax = 1000
        # c_0 = 1
        # c_j = c_0 + (1 - c_0)  # add complexity here

        self.active_indices |= set(new_active_indices)
        example_count = len(self.active_indices)
        x_k_j = [data[i] for i in self.active_indices]
        y_k = [labels[i] for i in self.active_indices]
        # labels = [row[1] for row in new_active_indices]

        # Variables

        # Equation offsets
        b_i = [
            m.addVar(
                vtype=GRB.CONTINUOUS, lb=-cmax, ub=cmax, name="b_i({i})".format(i=i)
            )
            for i in range(n_c)
        ]

        # Equation parameters
        a_i_j = [
            [
                m.addVar(
                    vtype=GRB.CONTINUOUS,
                    lb=-wmax,
                    ub=wmax,
                    name="a_i_j({i},{j})".format(i=i, j=j),
                )
                for j in range(n_v)
            ]
            for i in range(n_c)
        ]

        # Equation variable occurrence
        occ_i_j = [
            [
                m.addVar(vtype=GRB.BINARY, name="occ_i_j({i},{j})".format(i=i, j=j))
                for j in range(n_v)
            ]
            for i in range(n_c)
        ]

        # Constraint coverage
        cov_i = [
            m.addVar(vtype=GRB.BINARY, name="cov_i({i})".format(i=i))
            for i in range(n_c)
        ]

        # Example x constraint coverage
        cov_k_i = [
            [
                m.addVar(vtype=GRB.BINARY, name="cov_k_i({k},{i})".format(k=k, i=i))
                for i in range(n_c)
            ]
            for k in range(example_count)
        ]

        # Minimize equation variable occurrence => TODO why not constraint occurrence?
        m.setObjective(
            quicksum(1 * occ_i_j[i][j] for i in range(n_c) for j in range(n_v)),
            GRB.MINIMIZE,
        )

        # Constraints

        # 1 sum(wij*xkj<=ci) A positives
        for k in range(example_count):
            if y_k[k]:
                for i in range(
                    n_c
                ):  # one per positive example and number of constraints
                    m.addConstr(
                        sum(a_i_j[i][j] * x_k_j[k][j] for j in range(n_v)) <= b_i[i],
                        name="c1({k},{i})".format(k=k, i=i),
                    )
        # 2 sum(wij*sjk>m*ski-M+ci+e) A negatives
        for k in range(example_count):
            if not y_k[k]:
                for i in range(
                    n_c
                ):  # one per positive example and number of constraints
                    m.addConstr(
                        sum(a_i_j[i][j] * x_k_j[k][j] for j in range(n_v))
                        >= bigm * cov_k_i[k][i] - bigm + b_i[i] + ep,
                        name="c2({k},{i})".format(k=k, i=i),
                    )
        # 3 sum(ski>=1) A negatives
        for k in range(example_count):  # one per negative example
            if not y_k[k]:
                m.addConstr(
                    sum(cov_k_i[k][i] for i in range(n_c)) >= 1,
                    name="c3({k})".format(k=k),
                )

        # 5wij<=cmax*wbij
        for i in range(n_c):
            for j in range(n_v):
                m.addConstr(
                    a_i_j[i][j] <= wmax * occ_i_j[i][j],
                    name="c5({i},{j})".format(i=i, j=j),
                )
        # 6wij>=-camxwbij
        for i in range(n_c):
            for j in range(n_v):
                m.addConstr(
                    a_i_j[i][j] >= -wmax * occ_i_j[i][j],
                    name="c6({i},{j})".format(i=i, j=j),
                )
        # 7ci<=cmaxcbi
        for i in range(n_c):
            m.addConstr(b_i[i] <= cmax * cov_i[i], name="c7({i})".format(i=i))
        # 8ci>=-cmaxcbi
        for i in range(n_c):
            m.addConstr(b_i[i] >= -cmax * cov_i[i], name="c8({i})".format(i=i))
        # 4 sum wbij>=cbi
        for i in range(n_c):
            m.addConstr(
                sum(occ_i_j[i][j] for j in range(n_v)) >= cov_i[i],
                name="c4({i})".format(i=i),
            )

        m.optimize()

        if m.status == GRB.Status.OPTIMAL:
            inequalitys = []
            for i in range(n_c):
                getvariables = [Symbol(domain.variables[j], REAL) for j in range(n_v)]
                rightside = Real(b_i[i].x)
                inequalitys.append(
                    GE(
                        rightside,
                        Plus(Real(a_i_j[i][j].x) * getvariables[j] for j in range(n_v)),
                    )
                )
            theory = And(t for t in inequalitys)

            return theory

        else:
            raise NoFormulaFound(data, labels)



from pywmi import Domain
from pysmt.shortcuts import (
    REAL,
    Or,
    And,
    LE,
    Real,
    Symbol,
    BOOL,
    GT,
    Not,
    Plus,
    Times,
    GE,
)


def xy_domain():
    variables = ["x", "y"]
    var_types = {"x": REAL, "y": REAL}
    var_domains = {"x": (0, 1), "y": (0, 1)}
    return Domain(variables, var_types, var_domains)


def simple_checker_problem():
    theory = Or(
        And(LE(Symbol("x", REAL), Real(0.5)), LE(Symbol("y", REAL), Real(0.5))),
        And(GT(Symbol("x", REAL), Real(0.5)), GT(Symbol("y", REAL), Real(0.5))),
    )

    return xy_domain(), theory, "simple_checker"


def simple_checker_problem_cnf():
    x, y = (Symbol(n, REAL) for n in ["x", "y"])
    theory = ((x <= 0.5) | (y > 0.5)) & ((x > 0.5) | (y <= 0.5))
    return xy_domain(), theory, "simple_cnf_checker"


def checker_problem():
    variables = ["x", "y", "a"]
    var_types = {"x": REAL, "y": REAL, "a": BOOL}
    var_domains = {"x": (0, 1), "y": (0, 1)}

    theory = Or(
        And(
            LE(Symbol("x", REAL), Real(0.5)),
            LE(Symbol("y", REAL), Real(0.5)),
            Symbol("a", BOOL),
        ),
        And(
            GT(Symbol("x", REAL), Real(0.5)),
            GT(Symbol("y", REAL), Real(0.5)),
            Symbol("a", BOOL),
        ),
        And(
            GT(Symbol("x", REAL), Real(0.5)),
            LE(Symbol("y", REAL), Real(0.5)),
            Not(Symbol("a", BOOL)),
        ),
        And(
            LE(Symbol("x", REAL), Real(0.5)),
            GT(Symbol("y", REAL), Real(0.5)),
            Not(Symbol("a", BOOL)),
        ),
    )

    return Domain(variables, var_types, var_domains), theory, "checker"


def simple_univariate_problem():
    variables = ["x"]
    var_types = {"x": REAL}
    var_domains = {"x": (0, 1)}

    theory = LE(Symbol("x", REAL), Real(0.6))

    return Domain(variables, var_types, var_domains), theory, "one_test"


def shared_hyperplane_problem():
    domain = xy_domain()
    x, y = (domain.get_symbol(v) for v in ["x", "y"])
    # y <= -x + 1.25
    shared1 = LE(y, Plus(Times(Real(-1.0), x), Real(1.25)))
    # y >= -x + 0.75
    shared2 = GE(y, Plus(Times(Real(-1.0), x), Real(0.75)))

    # y <= x + 0.5
    h1 = LE(y, Plus(x, Real(0.5)))
    # y >= x + 0.25
    h2 = GE(y, Plus(x, Real(0.25)))

    # y <= x - 0.25
    h3 = LE(y, Plus(x, Real(-0.25)))
    # y >= x - 0.5
    h4 = GE(y, Plus(x, Real(-0.5)))
    return (
        domain,
        Or(And(shared1, shared2, h1, h2), And(shared1, shared2, h3, h4)),
        "shared",
    )


def cross_problem():
    domain = xy_domain()
    x, y = (domain.get_symbol(v) for v in ["x", "y"])
    top = y <= 0.9
    middle_top = y <= 0.7
    middle_bottom = y >= 0.5
    bottom = y >= 0.1

    left = x >= 0.2
    middle_left = x >= 0.4
    middle_right = x <= 0.6
    right = x <= 0.8
    theory = (top & middle_left & middle_right & bottom) | (
        left & middle_top & middle_bottom & right
    )
    return domain, theory, "cross"


def bool_xor_problem():
    variables = ["a", "b"]
    var_types = {"a": BOOL, "b": BOOL}
    var_domains = dict()
    domain = Domain(variables, var_types, var_domains)

    a, b = (domain.get_symbol(v) for v in variables)

    theory = (a & ~b) | (~a & b)
    return domain, theory, "2xor"


def ice_cream_problem():
    variables = ["chocolate", "banana", "weekend"]
    chocolate, banana, weekend = variables
    var_types = {chocolate: REAL, banana: REAL, weekend: BOOL}
    var_domains = {chocolate: (0, 1), banana: (0, 1)}
    domain = Domain(variables, var_types, var_domains)

    chocolate, banana, weekend = (domain.get_symbol(v) for v in variables)
    theory = (
        (chocolate < 0.650)
        & (banana < 0.550)
        & (chocolate + 0.7 * banana <= 0.700)
        & (chocolate + 1.2 * banana <= 0.750)
        & (~weekend | (chocolate + 0.7 * banana <= 0.340))
    )

    return domain, theory, "ice_cream"


def get_all():
    return [
        simple_checker_problem(),
        simple_checker_problem_cnf(),
        checker_problem(),
        simple_univariate_problem(),
        shared_hyperplane_problem(),
        cross_problem(),
        bool_xor_problem(),
        ice_cream_problem(),
    ]


def get_by_name(name):
    for t in get_all():
        if t[2] == name:
            return t[0], t[1]



import time

import os
from pywmi import evaluate, smt_to_nested

from incal.incremental_learner import IncrementalObserver

from pywmi.plot import plot_combined


class PlottingObserver(IncrementalObserver):
    def __init__(
        self,
        domain,
        directory,
        name,
        feat_x,
        feat_y,
        condition=None,
        auto_clean=False,
        run_name=None,
    ):
        self.domain = domain

        if not os.path.exists(directory):
            os.makedirs(directory)

        if auto_clean:
            run_number = 0
            run_dir = None
            while run_dir is None or os.path.exists(run_dir):
                date_folders = time.strftime("%Y{s}%m{s}%d{s}".format(s=os.path.sep))
                run_name = run_name + " " if run_name is not None else ""
                run_dir_name = "run {}{}".format(run_name, time.strftime("%Hh %Mm %Ss"))
                run_dir = os.path.join(directory, date_folders, run_dir_name)
                if run_number > 0:
                    run_dir += "_{}".format(run_number)
                run_number += 1
            os.makedirs(run_dir)
            directory = run_dir

        self.directory = directory

        self.name = name
        self.all_active = set()
        self.feat_x = feat_x
        self.feat_y = feat_y
        self.iteration = 0
        self.condition = condition

    def observe_initial(self, data, labels, initial_indices):
        self.all_active = self.all_active.union(initial_indices)
        name = "{}{}{}_{}".format(
            self.directory, os.path.sep, self.name, self.iteration
        )
        plot_combined(
            self.feat_x,
            self.feat_y,
            self.domain,
            None,
            (data, labels),
            None,
            name,
            initial_indices,
            set(),
            self.condition,
        )

    def observe_iteration(
        self, data, labels, formula, new_active_indices, solving_time, selection_time
    ):
        self.iteration += 1
        learned_labels = evaluate(self.domain, formula, data)
        name = "{}{}{}_{}".format(
            self.directory, os.path.sep, self.name, self.iteration
        )
        plot_combined(
            self.feat_x,
            self.feat_y,
            self.domain,
            formula,
            (data, labels),
            learned_labels,
            name,
            self.all_active,
            new_active_indices,
            condition=self.condition,
        )
        self.all_active = self.all_active.union(new_active_indices)



from pywmi import Domain

from lp.model import Model


def lp_domain(n, ranges=None):
    if ranges is None:
        ranges = [(None, None) for i in range(n)]
    return Domain.make([], ["x{}".format(i + 1) for i in range(n)], ranges)


def lp_2_6() -> Model:
    domain = lp_domain(2)
    x1, x2 = domain.get_symbols(domain.variables)
    return Model(
        domain,
        300 * x1 + 200 * x2,
        [2 * x1 + x2 <= 100, x1 + x2 <= 80, x1 <= 40, x1 >= 0, x2 >= 0],
        minimize=False,
        name="LP_2_6",
    )


def lp_2_7() -> Model:
    domain = lp_domain(4)
    x1, x2, x3, x4 = domain.get_symbols(domain.variables)
    return Model(
        domain,
        320 * x1 + 400 * x2 + 480 * x3 + 560 * x4,
        [
            0.06 * x1 + 0.03 * x2 + 0.02 * x3 + 0.01 * x4 >= 3.5,
            0.03 * x1 + 0.02 * x2 + 0.05 * x3 + 0.06 * x4 <= 3,
            0.08 * x1 + 0.03 * x2 + 0.02 * x3 + 0.01 * x4 == 4,
            x1 + x2 + x3 + x4 == 110,
        ]
        + [x >= 0 for x in domain.get_symbols(domain.variables)],
        minimize=False,
        name="LP_2_7",
    )


def lp_2_8() -> Model:
    domain = lp_domain(3)
    x1, x2, x3 = domain.get_symbols(domain.variables)
    return Model(
        domain,
        5 * x1 + 4 * x2 + 3 * x3,
        [
            2 * x1 + 3 * x2 + x3 <= 5,
            4 * x1 + x2 + 2 * x3 <= 11,
            3 * x1 + 4 * x2 + 2 * x3 <= 5,
        ]
        + [x >= 0 for x in domain.get_symbols(domain.variables)],
        minimize=False,
        name="LP_2_8",
    )


def lp_2_9() -> Model:
    domain = lp_domain(4)
    x1, x2, x3, x4 = domain.get_symbols(domain.variables)
    return Model(
        domain,
        3 * x1 - x2,
        [
            0 - x1 + 6 * x2 - x3 + x4 >= -3,
            7 * x2 + 2 * x4 == 5,
            x1 + x2 + x3 - x4 <= 2,
            x1 >= 0,
            x3 >= 0,
        ],
        minimize=True,
        name="LP_2_9",
    )



import time

import pysmt.shortcuts as smt
from pysmt.exceptions import InternalSolverError

from .observe import observe
from .learner import Learner, NoFormulaFound


class IncrementalObserver(observe.SpecializedObserver):
    def observe_initial(self, data, labels, initial_indices):
        raise NotImplementedError()

    def observe_iteration(
        self, data, labels, formula, new_active_indices, solving_time, selection_time
    ):
        raise NotImplementedError()


class IncrementalLearner(Learner):
    def __init__(self, name, selection_strategy, smt_solver=True):
        """
        Initializes a new incremental learner
        :param str name: The learner name
        :param SelectionStrategy selection_strategy: The selection strategy
        """
        Learner.__init__(self, "incremental_{}".format(name))
        self.selection_strategy = selection_strategy
        self.observer = observe.DispatchObserver()
        self.smt_solver = smt_solver

    def add_observer(self, observer):
        self.observer.add_observer(observer)

    def learn(self, domain, data, labels, initial_indices=None):
        if self.smt_solver:
            with smt.Solver() as solver:
                data, formula, labels = self.incremental_loop(
                    domain, data, labels, initial_indices, solver
                )
        else:
            data, formula, labels = self.incremental_loop(
                domain, data, labels, initial_indices, None
            )

        return data, labels, formula

    def incremental_loop(self, domain, data, labels, initial_indices, solver):
        active_indices = (
            list(range(len(data))) if initial_indices is None else initial_indices
        )
        all_active_indices = active_indices
        self.observer.observe("initial", data, labels, active_indices)
        formula = None
        while len(active_indices) > 0:
            solving_start = time.time()
            try:
                formula = self.learn_partial(
                    solver, domain, data, labels, active_indices
                )
            except InternalSolverError:
                raise NoFormulaFound(data, labels)
            except Exception as e:
                if "Z3Exception" in str(type(e)):
                    raise NoFormulaFound(data, labels)
                else:
                    raise e

            solving_time = time.time() - solving_start

            selection_start = time.time()
            data, labels, new_active_indices = self.selection_strategy.select_active(
                domain, data, labels, formula, all_active_indices
            )
            active_indices = list(new_active_indices)
            all_active_indices += active_indices
            selection_time = time.time() - selection_start
            self.observer.observe(
                "iteration",
                data,
                labels,
                formula,
                active_indices,
                solving_time,
                selection_time,
            )
        return data, formula, labels

    def learn_partial(self, solver, domain, data, labels, new_active_indices):
        raise NotImplementedError()



import heapq

import time
import logging

from .learner import NoFormulaFound

logger = logging.getLogger(__name__)


class ParameterFrontier(object):
    def __init__(self, cost_function):
        self.pq = []
        self.tried = set()
        self.cost_function = cost_function

    def push(self, k, h):
        if (k, h) not in self.tried:
            heapq.heappush(self.pq, (self.cost_function(k, h), k, h))
            self.tried.add((k, h))

    def pop(self):
        c, k, h = heapq.heappop(self.pq)
        return k, h


class SearchStrategy:
    def produce_initial(self):
        raise NotImplementedError()

    def produce_next(self, k, h):
        raise NotImplementedError()

    def cost(self, k, h):
        raise NotImplementedError()


class DoubleSearchStrategy(SearchStrategy):
    def __init__(self, w_k=1, w_h=1, init_k=1, init_h=0, max_k=None, max_h=None):
        self.w_k = w_k
        self.w_h = w_h
        self.init_k = init_k
        self.init_h = init_h
        self.max_k = max_k
        self.max_h = max_h

    def produce_initial(self):
        return self.init_k, self.init_h

    def produce_next(self, k, h):
        result = []
        if self.max_k is None or k + 1 <= self.max_k:
            result.append((k + 1, h))
        if self.max_h is None or h + 1 <= self.max_h:
            result.append((k, h + 1))
        return result

    def cost(self, k, h):
        return self.w_k * k + self.w_h * h


class LpSearchStrategy(SearchStrategy):
    def __init__(self, init_h=1, max_h=None):
        self.init_h = init_h
        self.max_h = max_h

    def produce_initial(self):
        return self.init_h, self.init_h

    def produce_next(self, k, h):
        return [] if (self.max_h is not None and h >= self.max_h) else [(k + 1, h + 1)]

    def cost(self, k, h):
        return h


def learn_bottom_up(data, labels, learn_f, search_strategy: SearchStrategy):
    """
    Learns a CNF(k, h) SMT formula phi using the learner encapsulated in init_learner such that
    C(k, h) = w_k * k + w_h * h is minimal.
    :param data: List of tuples of assignments and labels
    :param labels: Array of labels
    :param learn_f: Function called with data, k and h: learn_f(data, k, h)
    :param search_strategy: The search strategy
    :return: A tuple containing: 1) the CNF(k, h) formula phi with minimal complexity C(k, h); 2) k; and 3) h
    """
    solution = k = h = None
    frontier = ParameterFrontier(search_strategy.cost)
    frontier.push(*search_strategy.produce_initial())
    i = 0
    while solution is None:
        i += 1
        k, h = frontier.pop()
        logger.debug("Attempting to solve with k={} and h={}".format(k, h))
        start = time.time()
        try:
            solution = learn_f(data, labels, i, k, h)
            logger.debug("Found solution after {:.2f}s".format(time.time() - start))
        except NoFormulaFound as e:
            data = e.data
            labels = e.labels
        for next_node in search_strategy.produce_next(k, h):
            frontier.push(*next_node)
    return solution, k, h



def timeout(func, args=None, kwargs=None, duration=1, default=None):
    if args is None:
        args = ()
    if kwargs is None:
        kwargs = dict()

    if duration is None:
        return func(*args, **kwargs)

    import signal

    class TimeoutError(Exception):
        pass

    def handler(signum, frame):
        raise TimeoutError()

    # set the timeout handler
    signal.signal(signal.SIGALRM, handler)
    signal.alarm(duration)

    try:
        result = func(*args, **kwargs)
    except TimeoutError:
        result = default
    finally:
        signal.alarm(0)

    return result



from __future__ import print_function

import numpy as np
import pysmt.shortcuts as smt
from pysmt.fnode import FNode
from pysmt.typing import REAL
from typing import Set, Tuple, List

from .incremental_learner import IncrementalLearner
from pywmi import Domain


class KCnfSmtLearner(IncrementalLearner):
    def __init__(
        self,
        conjunction_count,
        half_space_count,
        selection_strategy,
        symmetries,
        allow_negations=True,
    ):
        IncrementalLearner.__init__(self, "cnf_smt", selection_strategy)
        self.conjunction_count = conjunction_count
        self.half_space_count = half_space_count
        self.symmetries = symmetries
        self.allow_negations = allow_negations

    def learn_partial(
        self,
        solver,
        domain: Domain,
        data: np.ndarray,
        labels: np.ndarray,
        new_active_indices: Set,
    ):

        # Constants
        n_b_original = len(domain.bool_vars)
        n_b = n_b_original * 2
        n_r = len(domain.real_vars)

        n_h_original = self.half_space_count if n_r > 0 else 0
        n_h = n_h_original * 2 if self.allow_negations else n_h_original

        n_c = self.conjunction_count
        n_d = data.shape[0]

        real_indices = np.array(
            [domain.var_types[v] == smt.REAL for v in domain.variables]
        )
        real_features = data[:, real_indices]
        bool_features = data[:, np.logical_not(real_indices)]

        # Variables
        a_hr = [
            [smt.Symbol("a_hr[{}][{}]".format(h, r), REAL) for r in range(n_r)]
            for h in range(n_h_original)
        ]
        if "1" in self.symmetries:
            b_h = [smt.Real(1.0) for _ in range(n_h_original)]
        else:
            b_h = [smt.Symbol("b_h[{}]".format(h), REAL) for h in range(n_h_original)]

        s_ch = [
            [smt.Symbol("s_ch[{}][{}]".format(c, h)) for h in range(n_h)]
            for c in range(n_c)
        ]
        s_cb = [
            [smt.Symbol("s_cb[{}][{}]".format(c, b)) for b in range(n_b)]
            for c in range(n_c)
        ]

        # Aux variables
        s_ih = [
            [smt.Symbol("s_ih[{}][{}]".format(i, h)) for h in range(n_h)]
            for i in range(n_d)
        ]
        s_ic = [
            [smt.Symbol("s_ic[{}][{}]".format(i, c)) for c in range(n_c)]
            for i in range(n_d)
        ]

        def pair(real: bool, c: int, index: int) -> Tuple[FNode, FNode]:
            if real:
                return s_ch[c][index], s_ch[c][index + n_h_original]
            else:
                return s_cb[c][index], s_cb[c][index + n_b_original]

        def order_equal(pair1, pair2):
            x_t, x_f, y_t, y_f = pair1 + pair2
            return smt.Iff(x_t, y_t) & smt.Iff(x_f, y_f)

        def order_geq(pair1, pair2):
            x_t, x_f, y_t, y_f = pair1 + pair2
            return x_t | y_f | ((~x_f) & (~y_t))

        def pairs(c: int) -> List[Tuple[FNode, FNode]]:
            return [pair(True, c, i) for i in range(n_h_original)] + [
                pair(False, c, i) for i in range(n_b_original)
            ]

        def order_geq_lex(c1: int, c2: int):
            pairs_c1, pairs_c2 = pairs(c1), pairs(c2)
            assert len(pairs_c1) == len(pairs_c2)
            constraints = smt.TRUE()
            for j in range(len(pairs_c1)):
                condition = smt.TRUE()
                for i in range(j):
                    condition &= order_equal(pairs_c1[i], pairs_c2[i])
                constraints &= smt.Implies(
                    condition, order_geq(pairs_c1[j], pairs_c2[j])
                )
            return constraints

        # Constraints
        for i in new_active_indices:
            x_r, x_b, label = (
                [float(val) for val in real_features[i]],
                bool_features[i],
                labels[i],
            )

            for h in range(n_h_original):
                sum_coefficients = smt.Plus(
                    [a_hr[h][r] * smt.Real(x_r[r]) for r in range(n_r)]
                )
                solver.add_assertion(smt.Iff(s_ih[i][h], sum_coefficients <= b_h[h]))

            for h in range(n_h_original, n_h):
                solver.add_assertion(smt.Iff(s_ih[i][h], ~s_ih[i][h - n_h_original]))

            for c in range(n_c):
                solver.add_assertion(
                    smt.Iff(
                        s_ic[i][c],
                        smt.Or(
                            [smt.FALSE()]
                            + [(s_ch[c][h] & s_ih[i][h]) for h in range(n_h)]
                            + [s_cb[c][b] for b in range(n_b_original) if x_b[b]]
                            + [
                                s_cb[c][b]
                                for b in range(n_b_original, n_b)
                                if not x_b[b - n_b_original]
                            ]
                        ),
                    )
                )

            # --- [start] symmetry breaking ---
            # Mutually exclusive
            if "m" in self.symmetries:
                for c in range(n_c):
                    for h in range(n_h_original):
                        solver.add_assertion(~(s_ch[c][h] & s_ch[c][h + n_h_original]))
                    for b in range(n_b_original):
                        solver.add_assertion(~(s_cb[c][b] & s_cb[c][b + n_b_original]))

            # Normalized
            if "n" in self.symmetries:
                for h in range(n_h_original):
                    solver.add_assertion(
                        smt.Equals(b_h[h], smt.Real(1.0))
                        | smt.Equals(b_h[h], smt.Real(0.0))
                    )

            # Vertical symmetries
            if "v" in self.symmetries:
                for c in range(n_c - 1):
                    solver.add_assertion(order_geq_lex(c, c + 1))

            # Horizontal symmetries
            if "h" in self.symmetries:
                for h in range(n_h_original - 1):
                    solver.add_assertion(a_hr[h][0] >= a_hr[h + 1][0])
            # --- [end] symmetry breaking ---

            if label:
                solver.add_assertion(smt.And([s_ic[i][c] for c in range(n_c)]))
            else:
                solver.add_assertion(smt.Or([~s_ic[i][c] for c in range(n_c)]))

        solver.solve()
        model = solver.get_model()

        x_vars = [domain.get_symbol(domain.real_vars[r]) for r in range(n_r)]
        half_spaces = [
            smt.Plus([model.get_value(a_hr[h][r]) * x_vars[r] for r in range(n_r)])
            <= model.get_value(b_h[h])
            for h in range(n_h_original)
        ] + [
            smt.Plus([model.get_value(a_hr[h][r]) * x_vars[r] for r in range(n_r)])
            > model.get_value(b_h[h])
            for h in range(n_h - n_h_original)
        ]

        b_vars = [domain.get_symbol(domain.bool_vars[b]) for b in range(n_b_original)]
        bool_literals = [b_vars[b] for b in range(n_b_original)]
        bool_literals += [~b_vars[b] for b in range(n_b - n_b_original)]

        conjunctions = [
            [half_spaces[h] for h in range(n_h) if model.get_py_value(s_ch[c][h])]
            + [bool_literals[b] for b in range(n_b) if model.get_py_value(s_cb[c][b])]
            for c in range(n_c)
        ]

        return smt.And([smt.Or(conjunction) for conjunction in conjunctions])



from __future__ import print_function

import numpy as np
import scipy
from pysmt.exceptions import InternalSolverError
from pysmt.environment import Environment
from pysmt.shortcuts import TRUE
from pysmt.typing import REAL, BOOL
from pywmi import Domain, evaluate
from pywmi.sample import uniform
from typing import Dict, Any, Tuple, TYPE_CHECKING

from .core import SelectionStrategy

if TYPE_CHECKING:
    from pysmt.fnode import FNode


class OneClassStrategy(SelectionStrategy):
    def __init__(
        self,
        regular_strategy,
        thresholds,
        tight_mode=True,
        class_label=True,
        background_knowledge=None,
    ):
        self.regular_strategy = regular_strategy  # type: SelectionStrategy
        self.thresholds = thresholds
        self.tight_mode = tight_mode
        assert class_label, "Currently only the positive setting is supported"
        self.class_label = class_label
        self.environment = Environment()
        if background_knowledge is None:
            self.background_knowledge = self.environment.formula_manager.TRUE()
        else:
            self.background_knowledge = self.environment.formula_manager.normalize(
                background_knowledge
            )

    def find_violating(self, domain, data, labels, formula):
        fm = self.environment.formula_manager
        formula = fm.normalize(formula)
        real_symbols = {name: fm.Symbol(name, REAL) for name in domain.real_vars}
        bool_symbols = {name: fm.Symbol(name, BOOL) for name in domain.bool_vars}
        symbols = real_symbols.copy()
        symbols.update(bool_symbols)
        bounds = domain.get_bounds(fm)
        try:
            with self.environment.factory.Solver() as solver:
                solver.add_assertion(formula)
                solver.add_assertion(bounds)
                solver.add_assertion(self.background_knowledge)
                # equalities = []
                # for row, label in data:
                #     for real_var in domain.real_vars:
                #         sym = real_symbols[real_var]
                #         val = fm.Real(row[real_var])
                #         t = fm.Real(self.thresholds[real_var])
                #         equalities.append(fm.Ite(sym >= val, fm.Equals(sym - val, t), fm.Equals(val - sym, t)))
                # solver.add_assertion(fm.Or(*equalities))
                for i in range(len(labels)):
                    row = {v: data[i, j].item() for j, v in enumerate(domain.variables)}
                    label = labels[i] == 1
                    if label == self.class_label:
                        constraint = fm.Implies(
                            fm.And(
                                *[
                                    fm.Iff(
                                        bool_symbols[bool_var],
                                        fm.Bool(row[bool_var] == 1),
                                    )
                                    for bool_var in domain.bool_vars
                                ]
                            ),
                            fm.Or(
                                *[
                                    fm.Ite(
                                        real_symbols[real_var]
                                        >= fm.Real(row[real_var]),
                                        real_symbols[real_var] - fm.Real(row[real_var])
                                        >= fm.Real(self.thresholds[real_var]),
                                        fm.Real(row[real_var]) - real_symbols[real_var]
                                        >= fm.Real(self.thresholds[real_var]),
                                    )
                                    for real_var in domain.real_vars
                                ]
                            ),
                        )
                    elif label == (not self.class_label):
                        constraint = fm.Implies(
                            fm.And(
                                *[
                                    fm.Iff(
                                        bool_symbols[bool_var],
                                        fm.Bool(row[bool_var] == 1),
                                    )
                                    for bool_var in domain.bool_vars
                                ]
                            ),
                            fm.Or(
                                *[
                                    fm.Ite(
                                        real_symbols[real_var]
                                        >= fm.Real(row[real_var]),
                                        real_symbols[real_var] - fm.Real(row[real_var])
                                        >= fm.Real(self.thresholds[real_var]),
                                        fm.Real(row[real_var]) - real_symbols[real_var]
                                        >= fm.Real(self.thresholds[real_var]),
                                    )
                                    for real_var in domain.real_vars
                                ]
                            ),
                        )
                    else:
                        raise ValueError("Unknown label l_{} = {}".format(i, label))
                    solver.add_assertion(constraint)
                solver.solve()
                model = solver.get_model()
                example = [
                    float(model.get_value(symbols[var]).constant_value())
                    for var in domain.variables
                ]
        except InternalSolverError:
            return None
        except Exception as e:
            if "Z3Exception" in str(type(e)):
                return None
            else:
                raise e

        return example

    def select_active(self, domain, data, labels, formula, active_indices):
        data, labels, selected = self.regular_strategy.select_active(
            domain, data, labels, formula, active_indices
        )
        if len(selected) > 0:
            return data, labels, selected
        else:
            example = self.find_violating(domain, data, labels, formula)
            if example is None:
                return data, labels, []
            data = np.vstack([data, example])
            labels = np.append(labels, np.array([0 if self.class_label else 1]))
            return data, labels, [len(labels) - 1]

    @staticmethod
    def add_negatives(
        domain,
        data,
        labels,
        thresholds,
        sample_count,
        background_knowledge=None,
        distance_measure=None,
    ):
        # type: (Domain, np.ndarray, np.ndarray, Dict, int, FNode, Any) -> Tuple[np.ndarray, np.ndarray]

        new_data = uniform(domain, sample_count)
        background_knowledge = background_knowledge or TRUE()
        supported_indices = evaluate(domain, background_knowledge, new_data)
        boolean_indices = [
            i for i, v in enumerate(domain.variables) if domain.is_bool(v)
        ]
        real_indices = [i for i, v in enumerate(domain.variables) if domain.is_real(v)]
        for j in range(new_data.shape[0]):
            valid_negative = True
            for i in range(data.shape[0]):
                # noinspection PyTypeChecker
                if labels[i] and all(
                    data[i, boolean_indices] == new_data[j, boolean_indices]
                ):
                    in_range = True
                    for ri, v in zip(real_indices, domain.real_vars):
                        if abs(data[i, ri] - new_data[j, ri]) > thresholds[v]:
                            in_range = False
                            break
                    valid_negative = valid_negative and (not in_range)
                    if not valid_negative:
                        break
            supported_indices[j] = supported_indices[j] and valid_negative
        new_data = new_data[supported_indices == 1, :]
        return np.concatenate([data, new_data], axis=0), np.concatenate(
            [labels, np.zeros(new_data.shape[0])]
        )


"""
There is a point e, such that for every example e': d(e, e') > t
AND(OR(d(e, e', r) > t, forall r), forall e)
"""



import numpy as np
import pysmt.shortcuts as smt
from typing import Tuple

from pysmt.fnode import FNode
from pywmi import Domain


class NoFormulaFound(RuntimeError):
    def __init__(self, data, labels):
        self.data = data
        self.labels = labels


class Learner(object):
    def __init__(self, name):
        self.name = name

    def learn(
        self, domain: Domain, data: np.ndarray, labels: np.ndarray, border_indices
    ) -> Tuple[np.ndarray, np.ndarray, FNode]:
        raise NotImplementedError()

    @staticmethod
    def _convert(value):
        return float(value.constant_value())

    @staticmethod
    def _get_misclassification(data):
        true_count = 0
        for _, l in data:
            if l:
                true_count += 1
        return min(true_count, len(data) - true_count)

    @staticmethod
    def check_example(domain, example_features, dnf_list):
        x_vars = [domain.get_symbol(var) for var in domain.real_vars]
        b_vars = [domain.get_symbol(var) for var in domain.bool_vars]

        formula = smt.Or(
            [smt.And(hyperplane_conjunct) for hyperplane_conjunct in dnf_list]
        )
        substitution = {var: example_features[str(var)] for var in x_vars + b_vars}
        return formula.substitute(substitution).simplify().is_true()

    @staticmethod
    def fit_hyperplane(domain, examples):
        matrix = examples[:, [domain.is_real(v) for v in domain.variables]]
        k = np.ones((len(examples), 1))
        a = np.matrix.dot(np.linalg.inv(matrix), k)
        return a, 1



import os
import shutil
import sys

from setuptools import setup, find_packages, Command
from os import path

# To upload:
# pip install --upgrade twine wheel setuptools
# python setup.py upload

NAME = "incal"
DESCRIPTION = "Learning SMT(LRA) formulas"
URL = "https://github.com/smtlearning/incal"
EMAIL = "samuel.kolb@me.com"
AUTHOR = "Samuel Kolb"
REQUIRES_PYTHON = ">=3.5.0"
VERSION = "1.0.2"

# What packages are required for this module to be executed?
REQUIRED = [
    "pywmi",
    "numpy",
    "typing",
    "pysmt",
    "matplotlib",
    "scikit-learn",
]

# What packages are optional?
EXTRAS = {}

# Distribute: python setup.py upload
# Requires: pip install twine wheel

here = os.path.abspath(os.path.dirname(__file__))

with open(path.join(here, "README.md")) as ref:
    long_description = ref.read()


class UploadCommand(Command):
    """Support setup.py upload."""

    description = "Build and publish the package."
    user_options = []

    @staticmethod
    def status(s):
        """Prints things in bold."""
        print("\033[1m{0}\033[0m".format(s))

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        try:
            self.status("Removing previous builds…")
            shutil.rmtree(os.path.join(here, "dist"))
        except OSError:
            pass

        self.status("Building Source and Wheel (universal) distribution…")
        os.system("{0} setup.py sdist bdist_wheel --universal".format(sys.executable))

        self.status("Uploading the package to PyPI via Twine…")
        os.system("twine upload dist/*")

        # self.status('Pushing git tags…')
        # os.system('git tag v{0}'.format(about['__version__']))
        # os.system('git push --tags')

        sys.exit()


setup(
    name=NAME,
    version=VERSION,
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type="text/markdown",
    url=URL,
    author=AUTHOR,
    author_email=EMAIL,
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
    ],
    python_requires=REQUIRES_PYTHON,
    packages=find_packages(exclude=("tests",)),
    zip_safe=False,
    install_requires=REQUIRED,
    extras_require=EXTRAS,
    setup_requires=["pytest-runner"],
    tests_require=["pytest"],
    entry_points={
        "console_scripts": [
            "incal-experiments = incal.experiments.cli:main",
            "incal-track = incal.experiments.learn:track",
        ]
    },
    cmdclass={"upload": UploadCommand},
)



import random

from pysmt.shortcuts import Real

from typing import Tuple, Optional, Callable, List

import numpy as np
from pysmt.fnode import FNode
from pywmi.domain import Density, Domain

from incal.incremental_learner import IncrementalLearner
from incal.observe.inc_logging import LoggingObserver
from incal.parameter_free_learner import (
    learn_bottom_up,
    SearchStrategy,
    DoubleSearchStrategy,
    LpSearchStrategy,
)
from incal.violations.core import SelectionStrategy, RandomViolationsStrategy


class Formula(Density):
    def __init__(self, domain, support):
        super().__init__(domain, support, Real(1))

    @classmethod
    def from_state(cls, state: dict):
        density = Density.from_state(state)
        return cls(density.domain, density.support)


Indices = List[int]


def learn_incremental(
    domain: Domain,
    data: np.ndarray,
    labels: np.ndarray,
    learner_factory: Callable[[int, int, SelectionStrategy], IncrementalLearner],
    initial_strategy: Callable[[Indices], Indices],
    selection_strategy: SelectionStrategy,
    search_strategy: SearchStrategy,
    log: Optional[str] = None,
) -> Tuple[FNode, int, int]:
    """
    Learn a formula that separates the positive and negative examples
    :return: A tuple containing 1. the learned formula, 2. the number of terms (or clauses) used,
    3. the number of hyperplanes used
    """

    def learn_inc(_data, _labels, _i, _k, _h):
        iteration_learner = learner_factory(_k, _h, selection_strategy)
        initial_indices = initial_strategy(list(range(len(_data))))
        # log_file = os.path.join(log_dir, "{}_{}_{}.txt".format(problem_name, _k, _h))
        if log is not None:
            iteration_learner.add_observer(
                LoggingObserver(log, _k, _h, None, False, selection_strategy)
            )
        return iteration_learner.learn(domain, _data, _labels, initial_indices)

    ((_d, _l, formula), k, h) = learn_bottom_up(
        data, labels, learn_inc, search_strategy
    )
    return formula, k, h


def incal(
    domain: Domain,
    data: np.ndarray,
    labels: np.ndarray,
) -> Tuple[FNode, int, int]:
    from incal.k_cnf_smt_learner import KCnfSmtLearner

    return learn_incremental(
        domain,
        data,
        labels,
        lambda k, h, ss: KCnfSmtLearner(k, h, ss, ""),
        lambda ind: random.sample(ind, 20),
        RandomViolationsStrategy(10),
        DoubleSearchStrategy(1.4, 1),
    )


def incalp(
    domain: Domain,
    data: np.ndarray,
    labels: np.ndarray,
) -> Tuple[FNode, int]:
    from incal.lp_learner_milp import LpLearnerMilp

    formula, k, j = learn_incremental(
        domain,
        data,
        labels,
        lambda k, h, ss: LpLearnerMilp(k, h, ss),
        lambda ind: random.sample(ind, 20),
        RandomViolationsStrategy(10),
        LpSearchStrategy(),
    )
    return formula, k
